

#' SHAFF: SHApley eFfects via random Forests
#' 
#' Run shaff to estimate Shapley effects from input data.
#'
#' @param data Training data of class \code{data.frame}, \code{matrix}, \code{dgCMatrix} (Matrix) or \code{gwaa.data} (GenABEL). Name of output variable must be Y.
#' @param K Number of variable subsets U for the Importance Sampling. Default is 500.
#' @param num.trees Number of trees.
#' @param mtry Number of variables to possibly split at in each node. Default is the (rounded down) square root of the number variables. Alternatively, a single argument function returning an integer, given the number of independent variables.
#'
#' @return Vector of Shapley effects estimated by SHAFF.
#' @export
#'
shaff <- function(data, num.trees = 500, mtry = NULL, K = 500){
  
  # grow forest
  nX <- ncol(data) - 1
  forest <- shaff_forest(Y ~., data = data, shapley_depth = nX, tree_depth = 100, num_varU = K, mtry = mtry,
                         max.depth = 100, num.trees = num.trees, num.threads = 1, importance = 'shapley')
  
  Uall <- forest$set.varU.MC
  Vall <- 1 - forest$pred.varU
  VallComp <- 1 - forest$pred.varU.comp
  pUAll <- forest$pU
  numU <- forest$numU
  
  idNA <- is.na(Vall) | is.na(VallComp)
  Vall <- Vall[!idNA]
  VallComp <- VallComp[!idNA]
  Uall <- Uall[!idNA]
  pUAll <- pUAll[!idNA]
  numU <- numU[!idNA]
  
  prob <- pUAll
  prob <- prob / sum(prob)
  
  ### solve least-squared
  Z <- t(sapply(Uall, function(S){
    z <- rep(0, nX)
    z[S + 1] <- 1
    z
  }))
  Zcomp <- t(sapply(Uall, function(S){
    z <- rep(1, nX)
    z[S + 1] <- 0
    z
  }))
  Z <- rbind(Z, Zcomp)
  w <- apply(Z, 1, function(z){
    k <- sum(z)
    (nX - 1)/(choose(nX, k) * k * (nX - k))
  })
  w[w == Inf] <- 1000
  w <- w*numU/(num_varU*prob)
  
  ## add null and full U sets
  Z <- rbind(rep(0, nX), rep(1, nX), Z)
  Vall <- c(0, forest$r.squared, Vall, VallComp)
  w <- c(10000, 10000, w)
  
  W <- diag(sqrt(w))
  shap.lm <- nnls(A = W %*% Z, b = W %*% Vall)
  
  return(shap.lm$x)
  
}
